/*
	Copyright 2000, Chris Gonnerman

	Universal LPR (ulpr)
	lpd_print.c -- lpd compatible backend for ulpr

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License version 2.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
	for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc., 59
	Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
	Usage:
		lpd_print hostname printer

	if hostname is null string or a single period, 
		uses the default of localhost
	if printer is null string or a single period,
		uses the default of lp

	The standard input is read and sent to the indicated printer.

	This program is intended to be hidden behind the ulpr
	front-end program.
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>

#include "socket.h"

#define BLOCKSIZE	2048

int debug = 0;

int stdin_to_temp(void);


static int lpd_bind(int sock, struct sockaddr_in *localaddr)
{
	int trys_left, port_index;
	char *mode;

	if(debug) fprintf(stderr, "lpd_bind(sock=%d, localaddr=%lx)\n", 
		sock, (unsigned long)localaddr);

	mode = getenv("ULPRstrict");

	if(mode == NULL) /* just need a reserved port */
		return bind_resv(sock, localaddr);

	/* this is for strict rfc compliance */

	if(debug) fprintf(stderr, "lpd_bind -- strict mode selected.\n");

	trys_left = 11;
	port_index = time(NULL) % 11;

	while(trys_left > 0) {
		localaddr->sin_port = htons(721+port_index);
		if(bind(sock, (struct sockaddr *)localaddr, 
			sizeof(struct sockaddr_in)) != -1)
				break;
		if(errno == EACCES) {
			fprintf(stderr, "lpd_print should be installed setuid root\n");
			return -1;
		}
		port_index = (port_index + 1) % 11;
		trys_left--;
	}

	if(trys_left <= 0) {
		errno = EAGAIN; /* overloaded for our purposes. */
		fprintf(stderr, 
			"lpd_print:  can't get rfc-strict reserved port (none available)\n");
		return -1;
	}

	return 0;
}


static int lpd_open(char *hostname)
{
	struct hostent *host;
	struct sockaddr_in localaddr, remoteaddr;
	struct servent *sp;
	int e, sock, port_index, trys_left;
	unsigned long inaddr;

	if(debug) fprintf(stderr, "lpd_open(hostname=%s)\n", hostname);

	/* initialization */

	memset((void *)&localaddr, 0, sizeof(struct sockaddr_in));
	localaddr.sin_family = AF_INET;
	localaddr.sin_addr.s_addr = htonl(INADDR_ANY);

	memset((void *)&remoteaddr, 0, sizeof(struct sockaddr_in));
	remoteaddr.sin_family = AF_INET;

	/* look up remote host */

	if((inaddr = inet_addr(hostname)) != INADDR_NONE) {
		memcpy((void *)&(remoteaddr.sin_addr), 
			(void *)&inaddr, sizeof(inaddr));
	} else {
		if((host = gethostbyname(hostname)) == NULL) {
			errno = 0; /* means look at h_errno */
			return -1;
		}
		memcpy((void *)&(remoteaddr.sin_addr),
			host->h_addr_list[0], host->h_length);
	}
		
	/* look up the remote port */

	if((sp = getservbyname("printer", "tcp")) == NULL)
		remoteaddr.sin_port = htons(515);
	else
		remoteaddr.sin_port = sp->s_port;

	/* choose a random source address.  repeat until well done. */

	if((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return -1;

	if(lpd_bind(sock, &localaddr) == -1) {
		close(sock);
		return -1;
	}

	setalarm(timeout);

	if(connect(sock, (struct sockaddr *)&remoteaddr, 
		sizeof(remoteaddr)) != -1) {
			alarm(0);
			return sock;
		}

	alarm(0);
	
	e = errno;

	fprintf(stderr, "lpd_print: can't open connection\n");

	close(sock);
	errno = e;
	return -1;
}


static void converse_error(int rc, char *hostname)
{
	if(debug) fprintf(stderr, "converse_error(rc=%d, hostname=%s)\n", 
		rc, hostname);

	if(rc == -1)
		fprintf(stderr, "lpd_print:  errno %d from server %s\n",
			errno, hostname);
	else
		fprintf(stderr, "lpd_print:  return code %d from server %s\n",
			rc, hostname);
}


static int lpd_converse(char *message, int len, int sock)
{
	char ch;

	if(debug) {
		ch = message[len-1];
		message[len-1] = '\0';
		fprintf(stderr, "lpd_converse(message=[%x]%s, len=%d, sock=%d)\n", 
			message[0], message+1, len, sock);
		message[len-1] = ch;
	}

	if(net_write(sock, message, len) == -1)
		return -1;

	if(net_read(sock, &ch, 1) == -1)
		return -1;
	
	return ch;
}


static int get_seqno()
{
	FILE *fp;
	int seq = 0;
	char buf[10+1];

	if(debug) fprintf(stderr, "get_seqno()\n");

	if((fp = fopen(SEQNOFILE, "r+")) == NULL)
		return -1;

	flock(fileno(fp), LOCK_EX);

	buf[10] = '\0';

	if(fgets(buf, 10, fp) == NULL) {
		seq = 0;
		strcpy(buf, "1\n");
	} else {
		seq = atoi(buf);
		sprintf(buf, "%3d\n", seq+1);
	} 

	rewind(fp);
	fputs(buf, fp);
	fflush(fp);

	flock(fileno(fp), LOCK_UN);

	fclose(fp);
	
	return seq;
}


static int push_job(int sock, int seq, char *hostname, char *queue)
{
	FILE *fp;
	struct stat stbuf;
	char buf[BLOCKSIZE], df_filename[128], my_hostname[128],
		*my_username, message[128];
	int len, ch, rc;
	size_t readsize;
	off_t nleft;
	char *filename;

	/* initialization */

	filename = "ulpr_doc";

	if((my_username = getenv("USER")) == NULL)
		my_username = "user";

	if(gethostname(my_hostname, 128) == -1) {
		fprintf(stderr, "lpd_print:  gethostname() failed.\n");
		return;
	}

	if(strcmp(hostname, ".") == 0)
		hostname = "localhost";

	if(strcmp(queue, ".") == 0)
		queue = "lp";

	/* give up priviledges */

	setuid(getuid());

	/*
		rfc 1179 says all servers must be
		able to accept control file first... 
		so open w-i-d-e, here it is.
	*/

	snprintf(df_filename, 128, "dfA%03d%s", seq, my_hostname);

	snprintf(buf, BLOCKSIZE, "H%s\nP%s\nJ%s\nC%s\nf%s\nU%s\nN%s\n",
		my_hostname, my_username, filename, filename,
		df_filename, df_filename, filename);

	len = strlen(buf);

	snprintf(message, 128, "\02%d cfA%03d%s\n", len, seq, my_hostname);

	if((rc = lpd_converse(message, strlen(message), sock)) != 0) {
		close(sock);
		converse_error(rc, hostname);
		fprintf(stderr, "            send control file failed.\n");
		return;
	}

	if(net_write(sock, buf, len) == -1) {
		close(sock);
		fprintf(stderr, "lpd_print:  errno %d writing control file.\n", 
			errno);
		return;
	}

	message[0] = '\0';

	if((rc = lpd_converse(message, 1, sock)) != 0) {
		close(sock);
		converse_error(rc, hostname);
		fprintf(stderr, "            control file not received.\n");
		return;
	}

	/* open the file */

	if(fstat(0, &stbuf) == -1) {
		fprintf(stderr, "lpd_print:  stat() failed, skipping.\n");
		return;
	}

	fp = stdin;

	if(!S_ISREG(stbuf.st_mode)) {
		FILE *ofp;

		/* not a regular file, need to create a temp file. */
		strcpy(buf, TMPDIR "/ulprXXXXXX");
		mktemp(buf);

		if((ofp = fopen(buf, "w+")) == NULL) {
			fprintf(stderr, 
				"lpd_print:  can't open temp file %s for read/write",
				buf);
			return;
		}

		unlink(buf);

		while((ch = getchar()) != EOF)
			putc(ch, ofp);

		fp = ofp;
		rewind(fp);

		if(fstat(fileno(fp), &stbuf) == -1) {
			fprintf(stderr, "lpd_print:  error, can't stat() my tempfile!\n");
			fclose(fp);
			return;
		}
	}

	/* now send the data file */

	snprintf(message, 128, "\03%ld dfA%03d%s\n", (long)stbuf.st_size, 
		seq, my_hostname);

	if((rc = lpd_converse(message, strlen(message), sock)) != 0) {
		close(sock);
		converse_error(rc, hostname);
		fprintf(stderr, "            send data file failed.\n");
		return;
	}

	nleft = stbuf.st_size;
	readsize = (nleft > BLOCKSIZE ? BLOCKSIZE : (int)nleft);

	while(nleft > 0 && (rc = fread(buf, 1, readsize, fp)) == readsize) {
		if(net_write(sock, buf, readsize) == -1) {
			fclose(fp);
			close(sock);
			fprintf(stderr, "lpd_print:  errno %d writing data file.\n", 
				errno);
			return;
		}

		nleft -= readsize;
		readsize = (nleft > BLOCKSIZE ? BLOCKSIZE : (int)nleft);
	}

	if(rc == 0) {
		fclose(fp);
		close(sock);
		fprintf(stderr, "lpd_print:  unexpected eof reading data file.\n");
		return;
	}

	if(getc(fp) != EOF) {
		fclose(fp);
		close(sock);
		fprintf(stderr, "lpd_print:  data file too long.\n");
		return;
	}

	fclose(fp);

	message[0] = '\0';

	if((rc = lpd_converse(message, 1, sock)) != 0) {
		close(sock);
		converse_error(rc, hostname);
		fprintf(stderr, "            data file not received.\n");
		return;
	}
}


/*
	send filename to queue at hostname

	note: creates one connection per file.  in strict mode,
	may use up outgoing ports quickly.
*/
static int lpd_send_job(char *hostname, char *queue)
{
	int status, sock, nfiles, i, seq, rc;
	char buf[BLOCKSIZE];
	struct stat stbuf;

	if(debug) 
		fprintf(stderr, "lpd_send_job(hostname=%s, queue=%s)\n",
			hostname, queue);

	/* stat the file */

	if(fstat(0, &stbuf) == -1) {
		fprintf(stderr, "lpd_print: can't stat() stdin\n");
		return -1;
	}

	/* open hailing frequencies */

	if((sock = lpd_open(hostname)) < 0) 
		return -1;

	/* request to send job */

	snprintf(buf, BLOCKSIZE, "\02%s\n", queue);

	if((rc = lpd_converse(buf, strlen(buf), sock)) != 0) {
		close(sock);
		converse_error(rc, hostname);
		fprintf(stderr, "            can't initiate job transfer.\n");
		return -1;
	}

	/* now in send-job mode */

	seq = get_seqno();

	push_job(sock, seq, hostname, queue);

	close(sock);

	return 0;
}


int main(int argc, char **argv)
{
	char *opt, *file;
	int i;

	if((opt = getenv("ULPRdebug")) != NULL)
		debug = 1;

	if((opt = getenv("ULPRtime")) != NULL) {
		timeout = atoi(opt);
		if(timeout < 1)
			timeout = 15;
	}

	if(argc != 3) {
		fprintf(stderr, "lpd_print:  invalid argument list;"
			" is the frontend broken?\n");
		exit(1);
	}

	if(lpd_send_job(argv[1], argv[2]) != 0)
		exit(1);

	exit(0);
}


/* end of file. */
