/*
	Copyright 2000, Chris Gonnerman

	Universal LPR (ulpr)
	socket.c -- generic socket I/O handlers

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License version 2.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
	for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc., 59
	Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>

extern int debug;	/* main procedure file must define */

int timeout = 60;	/* seconds */


int canread(int fd)
{
	fd_set r, e;
	struct timeval tv;
	int rc;

	if(debug) fprintf(stderr, "canread(fd=%d)\n", fd);

	FD_ZERO(&r);
	FD_ZERO(&e);

	FD_SET(fd, &r);
	FD_SET(fd, &e);

	tv.tv_sec = timeout;
	tv.tv_usec = 0;

	rc = select(fd+1, &r, NULL, &e, &tv);

	if(debug) {
		fprintf(stderr, "canread() -- select() returned %d\n", rc);
		fprintf(stderr, "          -- FD_ISSET(%d, &r) = %d\n", 
			fd, FD_ISSET(fd, &r));
		fprintf(stderr, "          -- FD_ISSET(%d, &e) = %d\n", 
			fd, FD_ISSET(fd, &e));
	}

	if(!rc || FD_ISSET(fd, &e))
		return 0;

	if(FD_ISSET(fd, &r))
		return 1;

	/* shouldn't reach here... */
	return 0;
}


/*
	write nbytes from ptr to fd

	returns -1 on error, 0 on success
*/
int net_write(int fd, char *ptr, int nbytes)
{
	int	nleft, nwritten;

	if(debug) fprintf(stderr, "net_write(fd=%d, ptr=%lx, nbytes=%d)\n", 
		fd, (unsigned long)ptr, nbytes);

	nleft = nbytes;

	while (nleft > 0) {

		nwritten = write(fd, ptr, nleft);

		if (nwritten <= 0) {
			if(debug) fprintf(stderr, "net_write() failed!\n");
			return -1;
		}

		nleft -= nwritten;
		ptr   += nwritten;
	}

	return 0;
}

/*
	read up to nbytes into ptr from fd
	
	returns actual bytes read, or -1 on error
*/
int net_read(int fd, char *ptr, int nbytes)
{
	int	nleft, nread;

	if(debug) fprintf(stderr, "net_read(fd=%d, ptr=%lx, nbytes=%d)\n", 
		fd, (unsigned long)ptr, nbytes);

	nleft = nbytes;
	while (nleft > 0) {
		if(!canread(fd)) {
			if(debug) fprintf(stderr, "timeout in net_read()\n");
			errno = ETIMEDOUT;
			return -1;
		}
		nread = read(fd, ptr, nleft);
		if (nread < 0) {
			if(debug) fprintf(stderr, "net_read() failed!\n");
			return(nread);		/* error, return < 0 */
		} else if (nread == 0)
			break;			/* EOF */

		nleft -= nread;
		ptr   += nread;
	}
	return(nbytes - nleft);		/* return >= 0 */
}


int bind_resv(int sock, struct sockaddr_in *localaddr)
{
	int port_index;

	for(port_index = IPPORT_RESERVED - 1; 
	    port_index > (IPPORT_RESERVED / 2);
		port_index--) 
	{
		localaddr->sin_port = htons(port_index);
		if(bind(sock, (struct sockaddr *)localaddr, 
			sizeof(struct sockaddr_in)) >= 0) {
				return 0;
		}
		if(errno == EACCES) {
			fprintf(stderr, 
				"socket:  should be installed setuid root\n");
			return -1;
		}
		if(errno != EADDRINUSE) {
			fprintf(stderr, 
				"socket:  unexpected errno %d binding port\n",
				errno);
			return -1;
		}
	}

	/* failed to get any reserved ports... */
	errno = EADDRINUSE;
	fprintf(stderr, 
		"socket:  can't get reserved port (none available)\n");
	return -1;
}


/* not specifically socket oriented, but ulpr backends use it a lot */


static void alarmhandler(int signal)
{
	/* doesn't do a darned thing */
	;
}


void setalarm(int timeout)
{
	struct sigaction sa;

	sigaction(SIGALRM, NULL, &sa);
	sa.sa_flags |= SA_ONESHOT;
	sa.sa_flags &= !(SA_RESTART);
	sa.sa_handler = alarmhandler;
	sigaction(SIGALRM, &sa, NULL);

	alarm(timeout);
}


/* end of file. */
