/*
	Copyright 2000, Chris Gonnerman

	Universal LPR (ulpr)
	tcp_print.c -- raw tcp backend for ulpr

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License version 2.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
	for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc., 59
	Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
	Usage:
		tcp_print hostname printer

	in this case, printer is actually the tcp port (as a number or
	symbolic name) to open and dump into.  each file printed is sent
	on a separate connection.

	if hostname is null string or a single period, 
		uses the default of localhost
	if printer is null string or a single period,
		uses the default of 10001

	The standard input is read and sent to the indicated printer.

	This program is intended to be hidden behind the ulpr
	front-end program.
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <time.h>

#include "socket.h"

#define BLOCKSIZE	2048

int debug = 0;
int reserved = 0;


static int tcp_open(char *hostname, char *portname)
{
	struct hostent *host;
	struct sockaddr_in remoteaddr;
	struct servent *sp;
	int e, sock, port_index, trys_left;
	unsigned long inaddr;

	if(debug) 
		fprintf(stderr, "tcp_open(hostname=%s, portname=%s)\n",
			hostname, portname);

	/* initialization */

	memset((void *)&remoteaddr, 0, sizeof(struct sockaddr_in));
	remoteaddr.sin_family = AF_INET;

	/* look up remote host */

	if((inaddr = inet_addr(hostname)) != INADDR_NONE) {
		memcpy((void *)&(remoteaddr.sin_addr), 
			(void *)&inaddr, sizeof(inaddr));
	} else {
		if((host = gethostbyname(hostname)) == NULL) {
			errno = 0; /* means look at h_errno */
			return -1;
		}
		memcpy((void *)&(remoteaddr.sin_addr),
			host->h_addr_list[0], host->h_length);
	}
		
	/* look up the remote port */

	if((sp = getservbyname(portname, "tcp")) == NULL)
		remoteaddr.sin_port = htons(atoi(portname));
	else
		remoteaddr.sin_port = sp->s_port;

	if(debug) 
		fprintf(stderr, "tcp_open: port number is %d\n", 
			ntohs(remoteaddr.sin_port));

	if((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		fprintf(stderr, "tcp_print:  can't create socket, errno = %d\n",
			errno);
		return -1;
	}

	if(reserved) {
		struct sockaddr_in localaddr;

		memset((void *)&localaddr, 0, sizeof(struct sockaddr_in));
		localaddr.sin_family = AF_INET;
		localaddr.sin_addr.s_addr = htonl(INADDR_ANY);

		if(bind_resv(sock, &localaddr) == -1) {
			fprintf(stderr, 
				"tcp_print:  reserved port requested, not available\n");
			return -1;
		}
	}

	setalarm(timeout);

	if(connect(sock, (struct sockaddr *)&remoteaddr, 
		sizeof(remoteaddr)) != -1) {
			alarm(0);
			return sock;
		}

	alarm(0);
	
	e = errno;

	fprintf(stderr, "tcp_print:  can't open connection, errno = %d\n", e);

	close(sock);
	errno = e;
	return -1;
}


static int tcp_send_job(char *hostname, char *portname)
{
	int sock, rc;
	char buf[BLOCKSIZE];

	if((sock = tcp_open(hostname, portname)) < 0) {
		fprintf(stderr, "tcp_print:  open error %d to %s:%s\n",
			errno, hostname, portname);
		return sock;
	}

	while((rc = fread(buf, 1, BLOCKSIZE, stdin)) > 0) {
		if(net_write(sock, buf, rc) == -1) {
			close(sock);
			fprintf(stderr, "tcp_print:  errno %d writing print job.\n", 
				errno);
			return -1;
		}
	}

	close(sock);

	if(ferror(stdin)) {
		fprintf(stderr, "tcp_print:  error reading data file.\n");
		return -1;
	}

	return 0;
}


int main(int argc, char **argv)
{
	char *opt, *file;
	char *hostname, *portname;
	int i;

	if((opt = getenv("ULPRdebug")) != NULL)
		debug = 1;

	if((opt = getenv("ULPRreserved")) != NULL)
		reserved = 1;

	if((opt = getenv("ULPRtime")) != NULL) {
		timeout = atoi(opt);
		if(timeout < 1)
			timeout = 15;
	}

	if(argc != 3) {
		fprintf(stderr, "tcp_print:  invalid argument list;"
			" is the frontend broken?\n");
		exit(1);
	}

	hostname = argv[1];
	portname = argv[2];

	if(strcmp(hostname, ".") == 0)
		hostname = "localhost";

	if(strcmp(portname, ".") == 0)
		portname = "10001";

	if(tcp_send_job(hostname, portname) != 0)
		exit(1);

	exit(0);
}


/* end of file. */
