/*
	Copyright 2000, Chris Gonnerman

	Universal LPR (ulpr)
	ulpr.c -- Universal LPR Main Program

	This program is free software; you can redistribute it and/or modify it
	under the terms of the GNU General Public License version 2.

	This program is distributed in the hope that it will be useful, but
	WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
	or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
	for more details.

	You should have received a copy of the GNU General Public License along
	with this program; if not, write to the Free Software Foundation, Inc., 59
	Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*
	Usage:

		ulpr [ -p | -P printer-uri ] [ files... ]

	Printer URI's may be in either of these forms:

		method:printer@hostname/options...
			or
		method://hostname/printer/options...

	These URI's are allowed to contain spaces, and are not url-encoded,
	so including spaces will usually require quoting from the shell.

	Parts may be omitted:

		printer@hostname	-- uses default method and options
		lpd:				-- uses default printer, hostname, and options
		                       for an lpd-type print job
		lp					-- uses default method, hostname, and options

	Generally, options may only be included if the rest of the URI is
	present; this may change in future version.

	A single period can replace a part to be omitted:

		lpd:.@remote.host	-- default lpd printer queue at remote.host
		smb://./printer		-- smb printer named "printer" at default host
	
	This program breaks down the URI and hands it to the backend pre-parsed.
	Defaults are provided by the backend, however.

	The backend will be spawned with one of the given files open as stdin
	and the URI parsed as args.  This means that a lot of processes will
	be spawned if a lot of files are printed; generally this should not
	be a problem, and it eliminates a security hassle with the lpd protocol.
*/


#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "ulpr.h"


static void usage()
{
	fputs("Usage:  ulpr [ -p | -P printer-uri ] [ files... ]\n", stderr);
	fputs("   or:  ulpr -h  (for more details)\n", stderr);

	exit(1);
}


static void run_backend(char *backend, 
	char *fakename, char *hostname, char *queue, int fd)
{
	int pid, status;

	switch(pid = fork()) {
	case -1:
		fprintf(stderr, "ulpr:  fatal, can't fork().\n");
		exit(9);
		break;
	case 0: /* child */
		close(0);
		dup(fd);
		close(fd);
		execl(backend, fakename, hostname, queue, NULL);
		break;
	default: /* parent */
		waitpid(pid, &status, 0);
		break;
	}
}


int main(int argc, char **argv)
{
	char *uri = "";
	int i, rc, fd;
	char *method, *hostname, *queue, *opts, *ename;
	char buf[MAXENVNAME+1];
	char backend[MAXPATHNAME+1];

	/* find the uri on the command line, if present */

	if(argc > 1) {
		if(strcasecmp(argv[1], "-p") == 0) {
			if(argc <= 2)
				usage();
			uri = argv[2];
			argc -= 2; 
			argv += 2;
		} else if(strncasecmp(argv[1], "-p", 2) == 0) {
			uri = argv[1] + 2;
			argc--;
			argv++;
		}
	}

	/* parse the uri */

	/* first, find the method, if there is one... */

	for(i = 0; uri[i] && isalnum(uri[i]); i++);

	if(i > 0 && uri[i] == ':') {
		/* we got a method! */
		method = uri;
		uri[i++] = '\0';
		uri += i;
	} else
		method = DEFMETHOD;

	/* next, figure out what type of uri it is */

	if(uri[0] == '/' && uri[1] == '/') {	/* url-type */

		uri += 2; /* skip the // */
		hostname = uri;

		for(i = 0; uri[i] && uri[i] != '/'; i++);	/* find the slash */

		if(uri[i] == '/') {
			uri[i++] = '\0';
			uri += i;
			queue = uri;
		} else
			queue = DEFQUEUE;

	} else {								/* mailto-type (assumed) */
		
		queue = uri;

		for(i = 0; uri[i] && uri[i] != '@'; i++);	/* find the 'at' */

		if(uri[i] == '@') {
			uri[i++] = '\0';
			uri += i;
			hostname = uri;
		} else
			hostname = DEFHOST;
	}

	/* search for options */
	for(i = 0; uri[i] && uri[i] != '/'; i++);	/* find the slash */

	if(uri[i] == '/') {
		uri[i++] = '\0';
		opts = uri + i;
	} else
		opts = "";

	/* clean up a bit */
	if(queue[0] == '\0')
		queue = DEFQUEUE;
	if(hostname[0] == '\0')
		hostname = DEFHOST;

	/* parse the options */

	strcpy(buf, "ULPR");
	buf[MAXENVNAME] = '\0';
	ename = buf + 4;
	
	while(opts[0] != '\0') {
		char *thisopt, *p, *value;

		/* extract the option */
		thisopt = opts;
		for(; *opts && *opts != '/'; opts++);
		if(*opts == '/')
			*opts++ = '\0';

		/* parse it */
		for(p = thisopt; *p && *p != '='; p++);

		if(*p == '\0') {
			value = "1";
		} else {
			value = p + 1;
			*p = '\0';
		}
			
		/* set it */
		strncpy(ename, thisopt, MAXENVNAME - 4);
		if(setenv(buf, value, 1) != 0)
			fprintf(stderr, "ulpr:  error saving option %s=%s\n", buf, value);
			
	}

	/* compose the backend name */

	backend[MAXPATHNAME] = '\0';
	strncpy(backend, BACKENDDIR "/", MAXPATHNAME);
	strncat(backend, method, MAXPATHNAME);
	strncat(backend, "_print", MAXPATHNAME);

	buf[MAXENVNAME] = '\0';
	strncpy(buf, method, MAXENVNAME);
	strncat(buf, "_print", MAXENVNAME);

	/* run the print job(s) */

	rc = 0;

	if(argc < 2) {
		run_backend(backend, buf, hostname, queue, 0);
	} else {
		for(i = 1; i < argc; i++) {
			if((fd = open(argv[i], O_RDONLY)) == -1) {
				fprintf(stderr, "ulpr:  can't open %s; skipping\n", argv[i]);
				rc = 1;
			} else {
				run_backend(backend, buf, hostname, queue, fd);
				close(fd);
			}
		}
	}

	exit(rc);
}


/* end of file. */
